



{{ **************************************************************************************
   *                                                                                    *     
   *  Stepper Motor Robot                                                               *
   *                                                                                    *
   *  Using 28BYJ48-12-300 Motor ULN2003 Motor Controller                               * 
   *                                                                                    *
   *  Each step is 5.626 degrees / 64 (gear reduction) or 0.087890625 degrees           *
   *  The coils are energized in 8 steps or 0.703125 degrees (8*0.087890625 degrees)    *
   *  Each revolution is 360 degrees/0.087890625 degrees/rev or 4096 steps or           *
   *  512 8-steps cycles.  This code uses a 4 step coil sequence which is slightly      *
   *  faster with less torque by skipping 1/2 steps.                                    *
   *                                                                                    *
   *  This code launches a control method for each motor in an independent cog.  The    *
   *  main code commands the motors by updating parameters of speed, distance, direction*
   *  distance to the target, and brakes.  The motor updates the main code and allows   *
   *  coordination by passing parameters back of remaining distance, remaining distance *
   *  to target and motor in motion.  The main code using a control loop to read        *
   *  sensors, determine status, define needed action and coordinate commands between   *
   *  motors.                                                                           *    
   *                                                                                    *
   *  Gregg Erickson - 2013 (MIT License)                                               *
   *                                                                                    *
   **************************************************************************************
}}



CON
        _clkmode = xtal1 + pll16x   ' Set Prop to Maximum Clock Speed
        _XinFREQ = 5_000_000

        TO_CM = 29_034              ' Microseconds to go 1 Centimeters

        PingPin=15                  ' Pin of Ping Sensor Signal
        LeftPin=4                   ' First Pin of 4-pin Sequence of Left Motor
        RightPin=0                  ' First Pin of 4-pin Sqquence of Right Motor

        LeftTurn=0                  ' Case Constants for Motor Coordination 
        RightTurn=1
        Straight=2
        RightTwist=3
        LeftTwist=4
        LeftCurve=5
        RightCurve=6

                                    ' Stepper Motor Sequence to Distance Conversion Ratios

        InchTicks=53                ' Motor Sequences to Move a Wheel 1 inch
        CentiTicks=21               ' Motor Sequences to Move a Wheel 1 Centimeter
        RevTicks=512                ' Motor Sequences to Rotate a Wheel 360 Degrees
        TwistDegTicks=3             ' Motor Sequences to Rotate 1 Degree Using 2 Wheels
        TurnDegTicks=6              ' Motor Sequences to Pivot 1 Degree Using 1 Wheel


        MailBoxDistance=26          ' Target Distance in Centimeters from End of Line
        RotationOffset=9            ' Offset of Sensor from End of Line after 180 Degree Rotation

VAR
                                

Long LeftStack[100]                ' Stacks for Cogs
Long RightStack[100]
Long PingStack[100]

Long LeftSpeed, RightSpeed                     ' Speed of Motor from -650 to 650
Long LeftDist, RightDist                       ' Distance, each step is 1/4 inches
Long RightLock,LeftLock                        ' In Motion, True or False
Long LeftBrake, RightBrake                     ' Brakes, 1=Energized, 0=coast
Long LeftTarget,RightTarget,TargetDistance     ' Distance to Target 
Long TurnRatio, TurnCount                      ' Wheel Ratio and Counter

Byte IRsensor,LastValid                        ' IR Sensor Reading and Previous Reading
Byte EndofLine,BeginOfLine,BeginFinal          ' Status of Line Detection for Special Cases

OBJ

Serial: "FullDuplexSerial"                      'Serial Object For Debugging


PUB Main |n, spd,range,angle,choice,selection,maximum,stepangle,average

'----------------------------- Start Serial Communications ----------------------------

serial.start(31,30,0,115_200)
serial.str(@starting)               ' Let User Know Program is Starting
serial.tx($0A)


'----------------------------- Initialize Variables and Pins---- ----------------------     
                              



TurnCount:=0                        ' Initial Zero for Rotation Counter
TurnRatio:=1                        ' Initial Ratio of Opposite Side Wheel Rotations During Turns

LastValid:=0                        ' Initial Last Valid Condition of IR Sensors, No Line Sensed
BeginFinal:=false                   ' Initial Condition is Beginning of Line Not Found
EndofLine:=false                    ' Initial Condition is End of Line Not Found

TargetDistance:=(MailBoxDistance-RotationOffset)*Centiticks   'Distance of Target After End of Line Found, In Centimeters with Offset After Rotation

                        

'----------------------------- Set Sensor Pins for Input and Indicator Pins for Output ------------


Dira[10..11]~                      ' Set Pins 10 & 11 to Input for IR Sensors
Dira[12..13]~~                     ' Set Pins 12 & 13 to Output for Indicator LEDs


'---------------------------- Start Stepper Motor Controls ------------------------------------

cognew(PingRange(PingPin,@Range),@PingStack)                                                     ' Launch Ping Sensor Loop in a Separate Cog
cognew(Motor(LeftPin,@LeftSpeed,@LeftDist,@LeftTarget,@LeftBrake,@LeftLock),@Leftstack)          ' Launch Motor Control Method in a Separate Cog
cognew(Motor(RightPin,@RightSpeed,@RightDist,@RightTarget,@RightBrake,@RightLock),@Rightstack)  
Autobrake(false)                                                                                 ' Verify Both Brakes are Off
Move(Straight,600,2000*CentiTicks,1)                                                             ' Launch in Straight line

'--------------------------------------------------------------------------------------------------------------
'--------------------------------------------------------------------------------------------------------------
'----------------------------  Line Following Command Loop ----------------------------------------------------
'--------------------------------------------------------------------------------------------------------------
'--------------------------------------------------------------------------------------------------------------
    
repeat
     
'------------- Read Sensors and Set Indicators  ------------------------------------------------------------------

   IRsensor:=ina[11..10]                                  ' Read IR Sensors

   outa[12..13]:=ina[10..11]                              ' Mirror IR Sensor on Indicator LEDs

'--------------- Debug Sensor Data -------------------------------------------------------------------------------
   
   serial.dec(IRSensor)
   serial.tx(" ")
    

'---------------- Begin and End of Line Special Conditions -------------------------------------------------------

  If Not(BeginOfLine) and IRSensor==0    ' If Line Not Found Yet (Sensors Zero), Override Sensor to Go Straight (Sensor Status = 3)
     IRSensor:=3
  Else                                   ' Once Line is Found (Sensors not zero), Stop Overriding Sensor and Remember Status
     BeginOfLine:=true
   
  if Not (EndofLine)                     ' If The End of Line Not Found, Travel to Target Unknown so Keep Resetting Target Distance
        LeftTarget:=1000
        RightTarget:=1000

'--------------------- Debug Data Dump -----------------------------------------------------------------------------
    
  serial.dec(LeftDist)
  serial.tx(" ")
  serial.dec(LeftTarget)
  serial.tx(" ") 
  serial.dec(RightDist)
  serial.tx(" ")
  serial.dec(RightTarget)
  serial.tx(" ")
  serial.dec(Range)
  serial.tx(" ")
  
  serial.tx($0D)
  serial.tx($0A)

'------------------ Change Motor Controls Based upon Sensor Readings --------------------------------------------


  if range>5                                             ' Proceed if No Objects Close
                                          
     Case IRsensor                                         ' Selection Move Based Upon IR Sensor

     
      0:                                                       ' Twist Towards Last Valid Direction if No Line
          TurnCount:=TurnCount+1                                    ' Increment Counter to Measure Turn Angle
                                                                     
          If LastValid==2
              Move(LeftTwist,300,110*TwistDegTicks,1)               
          Else 
              Move(RightTwist,300,110*TwistDegTicks,1)

          If TurnCount>450                                           ' Test Counter to Determine if 180 Degress at End of Course
              EndofLine:=true                                              ' If 180,  End is True
              BeginFinal:=true                                             '          Begin Final Target Countdown
       

      1:
          Move(RightCurve,600,3000,6000)                         ' Proportional Right Turn if Left Sensor Sees a Reflection from the Board
          LastValid:=IRsensor                                         ' Note This as Last Valid Sensor Reading
          TurnRatio:=(TurnRatio+1) <#6000
        
           

      2:
          Move(LeftCurve,600,3000,6000)                          ' Proportional Left Turn if the Right Sensor See a Reflection from the Board
          LastValid:=IRsensor                                          ' Note This as Last Valid Sensor Reading
          TurnRatio:=(TurnRatio+1)<#6000


      3:
      
         Move(Straight,600,3000,1)                                ' Straight Along Line if Both Sensors Don't See a Reflection from the Board
         Lastvalid:=IRsensor                                           ' Note This as Last Valid Sensor Reading
         TurnRatio:=1                                                  ' Reset Turn Ratio for Proportional Steerng
         TurnCount:=0                                                  ' Clear Turn Counter for Short Turns
         If EndOfLine 
             If BeginFinal                                             ' If Countdown Started Then Load Distance Once
                LeftTarget:=TargetDistance
                RightTarget:=TargetDistance
                BeginFinal:=false
        
      OTHER:
  Else 
      LeftDist:=0                                          ' Stop for Obstacle when Object Close
      RightDist:=0
   


'-------------------------- Motor Coordination Methods ------------------------------------------------


Pub Autobrake(_on)
   LeftBrake:=_On
   RightBrake:=_On


Pub Wait

 repeat while RightDist>0 or LeftDist>0


Pub Move(_Move,_Speed,_Distance,_Ratio)

  case _Move
    LeftTurn:
      RightSpeed:=_Speed
      RightDist:=_Distance
      LeftSpeed:=0
      LeftDist:=0
    RightTurn: ' Distance x 2 = Angle
      RightSpeed:=0
      RightDist:=0
      LeftSpeed:=_Speed
      LeftDist:=_Distance
    Straight:
      RightSpeed:=_Speed
      RightDist:=_Distance
      LeftSpeed:=_Speed
      LeftDist:=_Distance
    LeftTwist:
      RightSpeed:=_Speed
      RightDist:=_Distance
      LeftSpeed:=-_Speed/_Ratio
      LeftDist:=_Distance
    RightTwist: '
      RightSpeed:=-_Speed/_Ratio
      RightDist:=_Distance
      LeftSpeed:=_Speed
      LeftDist:=_Distance
    LeftCurve:
      RightSpeed:=_Speed
      RightDist:=_Distance
      LeftSpeed:=_Speed/_Ratio
      LeftDist:=_Distance/_Ratio
    RightCurve: '
      RightSpeed:=_Speed/_Ratio
      RightDist:=_Distance/_Ratio
      LeftSpeed:=_Speed
      LeftDist:=_Distance



'------------------------- Motor Control Method for Launching in Separate Cog ------------------------------------

Pub Motor(_pin,_SpeedPtr,_DistPtr,_TargetPtr,_BrakePtr,_LockPtr)| Q[8],P[8],i,j,k,_Speed,_Distance,_PDistance,StartPin, EndPin,BrakeOn,MaxPin

MaxPin:=7

' Stepper Sequence from Data Sheet

  p[0]:=%1100    ' 1
  p[1]:=%0110    ' 3
  p[2]:=%0011    ' 5
  p[3]:=%1001    ' 7

    ' Stepper Sequence from Data Sheet
  q[0]:=%1000
  q[1]:=%1100
  q[2]:=%0100
  q[3]:=%0110
  q[4]:=%0010
  q[5]:=%0011
  q[6]:=%0001
  q[7]:=%1001

  
Repeat
  'Receive Parameters
  long[_LockPtr]:=%1111_1111
  _Speed:=long[_SpeedPtr]

  ' Process Parameters and Implement
  If _speed>0
    StartPin:=_pin
    EndPin:=_pin+3
  Else
    StartPin:=_pin+3
    EndPin:=_pin
  dira[StartPin..EndPin]~~

  ||_speed
  _speed#>=0
  _speed<#=500

  If _Speed>0 and Long[_DistPtr]>0 and Long[_TargetPtr]>0

           repeat i from 0 to MaxPin

              If MaxPin==3
                outa[StartPin..EndPin]:=p[i]
                long[_LockPtr]:=p[i]
              Else
                outa[StartPin..EndPin]:=q[i]
                long[_LockPtr]:=q[i]               
             
             waitcnt(clkfreq/_speed+cnt)
           Long[_DistPtr]:=Long[_DistPtr]-1
           Long[_TargetPtr]:=Long[_TargetPtr]-1

  ' Clear Variables
  long[_LockPtr]:=0            

  outa[StartPin..EndPin]~
  if long[_BrakePtr]==true
     outa[StartPin..EndPin]~~


PUB PingRange(Pin,_RangePtr) : Microseconds | cnt1, cnt2
''Return Ping)))'s one-way ultrasonic travel time in microseconds
                                                                                 
Repeat
  outa[Pin]~                                                                    ' Clear I/O Pin
  dira[Pin]~~                                                                   ' Make Pin Output
  outa[Pin]~~                                                                   ' Set I/O Pin
  outa[Pin]~                                                                    ' Clear I/O Pin (> 2 s pulse)
  dira[Pin]~                                                                    ' Make I/O Pin Input
  waitpne(0, |< Pin, 0)                                                         ' Wait For Pin To Go HIGH
  cnt1 := cnt                                                                   ' Store Current Counter Value
  waitpeq(0, |< Pin, 0)                                                         ' Wait For Pin To Go LOW 
  cnt2 := cnt                                                                   ' Store New Counter Value
  Microseconds := (||(cnt1 - cnt2) / (clkfreq / 1_000_000)) >> 1                ' Return Time in s

  Long[_RangePtr]:=Microseconds* 10_00 / TO_CM                                  ' Return Distance In Millimeters
  
  waitcnt(clkfreq/10+cnt)

'---------------------- Data ----------------------------------


Dat

Starting  byte "Starting",0
Calc      byte "Calculating",0
Complete  byte "Complete",0
Selected  byte "Selected",0